import struct
import uuid

from gpt_image.geometry import Geometry
from gpt_image.partition import (Partition, PartitionAttribute,
                                 PartitionEntryArray)

PART_NAME = "test-part"


def test_partition_init_blank():
    part = Partition()
    # everything should be zeroed
    assert part.type_guid.data_bytes == b"\x00" * 16
    assert part.partition_guid.data_bytes == b"\x00" * 16
    assert part.first_lba.data_bytes == b"\x00" * 8
    assert part.last_lba.data_bytes == b"\x00" * 8
    assert part.attribute_flags.data_bytes == b"\x00" * 8
    assert part.partition_name.data_bytes == b"\x00" * 72


def test_partition_init_real():
    part = Partition(PART_NAME, 2 * 1024)
    assert type(uuid.UUID(bytes=part.type_guid.data)) is uuid.UUID
    assert type(uuid.UUID(bytes=part.partition_guid.data)) is uuid.UUID
    assert part.partition_name.data_bytes == struct.pack(
        "<72s", bytes(PART_NAME, encoding="utf_16_le")
    )
    assert part.partition_name.data == bytes(PART_NAME, encoding="utf_16_le")

    assert part.size == 2 * 1024

def test_attribute_set():
    part = Partition(PART_NAME, 2*1024)
    part.attribute_flags = PartitionAttribute.HIDDEN
    assert part.attribute_flags.data_bytes == (b"\x00" *7) + b"@"
    part.attribute_flags = PartitionAttribute.READ_ONLY
    assert part.attribute_flags.data_bytes == (b"\x00" *7) + b"P"

def test_attribute_clear():
    part = Partition(PART_NAME, 2*1024)
    part.attribute_flags = PartitionAttribute.HIDDEN
    assert part.attribute_flags.data_bytes == (b"\x00" *7) + b"@"
    part.attribute_flags = PartitionAttribute.NONE
    assert part.attribute_flags.data_bytes == b"\x00" * 8

def test_add():
    geo = Geometry(8 * 1024 * 1024)
    pa = PartitionEntryArray(geo)
    new_part = Partition(PART_NAME, 2 * 1024)
    pa.add(new_part)
    # name is encoded with utf_16_le
    assert PART_NAME == pa.entries[0].partition_name.data.decode(encoding="utf_16_le")


def test_unmarshal():
    part = Partition(PART_NAME, 4 * 1024)
    part_b = part.byte_structure

    part2 = Partition()
    part2.read(part_b)
    assert len(part2.partition_name.data_bytes) == part2.partition_name.length
    assert part2.partition_name.data == PART_NAME



from gpt_image.geometry import Geometry


def test_default_geometry():
    """Simple tests to ensure math is sane

    We use statically calculated values (magic numbers) instead of
    trusting more math :)

    """
    disk_size = 2 * 1024 * 1024
    geo = Geometry(disk_size)
    assert geo.total_sectors == 4096
    assert geo.total_lba == 4096
    assert geo.last_usable_lba == 4062
    assert geo.primary_header_byte == 512
    assert geo.primary_array_byte == 1024
    assert geo.alternate_lba == 4095
    assert geo.alternate_header_byte == 2096640
    assert geo.alternate_array_lba == 4063
    assert geo.alternate_array_byte == 2080256



import pytest
from gpt_image.disk import Disk
from gpt_image.partition import Partition

BYTE_DATA = b"\x01\x02\x03\x04"
DISK_SIZE = 2 * 1024 * 1024
OFFSET = 32


@pytest.fixture
def new_image(tmp_path):
    image_name = tmp_path / "test.img"
    disk = Disk(image_name)
    disk.create(DISK_SIZE * 2)
    part1 = Partition("partition1", 2 * 1024 * 1024)
    part2 = Partition("partition2", 3 * 1024 * 1024)
    disk.table.partitions.add(part1)
    disk.table.partitions.add(part2)
    disk.write()
    return image_name


def test_disk_create(tmp_path):
    """Test creating a new disk image"""
    disk_path = tmp_path / "test-disk.img"
    disk = Disk(disk_path)
    assert disk.image_path == disk_path.resolve()
    assert disk.sector_size == 512
    disk.create(DISK_SIZE)
    assert disk.table.primary_header.signature.data == b"EFI PART"
    assert disk.table.primary_header.backup is False


def test_disk_open(new_image):
    disk = Disk(new_image)
    disk.open()
    assert disk.size == DISK_SIZE * 2
    assert disk.table.primary_header.backup is False
    assert disk.table.secondary_header.backup is True
    assert disk.table.primary_header.partition_entry_lba.data == 2
    assert disk.table.secondary_header.partition_entry_lba.data == 8159
    assert len(disk.table.partitions.entries) == 2
    assert disk.table.partitions.entries[0].partition_name.data == "partition1"


# def test_write_data_default(new_image: Disk):
#     part = Partition("test-part", 1024, uuid.uuid4(), 1)
#     new_image.table.partitions.add(part)
#     new_image.update_table()
#     new_image.write_data(BYTE_DATA, part)
#     with open(new_image.image_path, "r+b") as f:
#         f.seek(part.first_lba.data * new_image.sector_size)
#         assert f.read(4) == BYTE_DATA
#
#
# def test_write_data_offset(new_image: Disk):
#     part = Partition("test-part", 1024, uuid.uuid4(), 1)
#     new_image.table.partitions.add(part)
#     new_image.update_table()
#     new_image.write_data(BYTE_DATA, part, OFFSET)
#     new_image.write_data(BYTE_DATA, part, OFFSET + OFFSET)
#     with open(new_image.image_path, "r+b") as f:
#         f.seek(part.first_lba.data * new_image.sector_size + OFFSET)
#         assert f.read(4) == BYTE_DATA
#         f.seek(part.first_lba.data * new_image.sector_size + OFFSET + OFFSET)
#         assert f.read(4) == BYTE_DATA
#
#
# def test_write_data_type_error(new_image: Disk):
#     part = Partition("test-part", 1024, uuid.uuid4())
#     with pytest.raises(ValueError):
#         new_image.write_data(str("foobar"), part)  # type: ignore



import struct
from typing import Union


class Entry:
    """Entry Class contains a GPT table entry data"""

    valid_types = [str, bytes, int]

    def __init__(self, offset: int, length: int, input_data: Union[str, bytes, int]):
        """Create a GPT table entry

        Args:
            offset: the location of the bytes relative to structure
            length: of entry in bytes
            input_data: value the field is set to. can be str, bytes, int
        """
        self.offset = offset
        self.length = length
        self.end = self.offset + self.length
        self.data = input_data

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value: Union[str, bytes, int]):
        if type(value) not in Entry.valid_types:
            raise ValueError(f"invalid entry type: {type(value)}")
        # @TODO: handle too large of integer
        if (isinstance(value, str) or isinstance(value, bytes)) and len(
            value
        ) > self.length:
            raise ValueError(f"insufficient space for entry data: {self.length}")
        self._data = value

    @property
    def data_bytes(self) -> bytes:
        if isinstance(self.data, int):
            return (self.data).to_bytes(self.length, "little")
        if isinstance(self.data, str):
            return struct.pack(
                f"<{str(self.length)}s", bytes(self.data, encoding="ascii")
            )
        return struct.pack(f"<{str(self.length)}s", self.data)



class Geometry:
    """Geometry of disk image

    This is a convenience class that provides geometry calculations for
    an underlying disk

    Attributes:
        sector_size: typically set to 512 bytes
        total_bytes: disk size in bytes
        total_sectors: number of sectors on the disk
        total_lba: number of logical blocks on the disk
        header_length: 92 bytes
        array_max_length: maximum partition array length in bytes (128*128)
        first_usable_lba: logical block where the partitions start
        last_usable_lba: logical block where the partitions end
        my_lba: logical block location of primary header
        primary_header_byte: byte where primary header starts
        partition_entry_lba: logical block where the primary partition array starts
        primary_array_byte: byte where the primary partition array starts
        alternate_lba: logical block where the backup header starts
        alternate_header_byte: byte where the backup header starts
        alternate_array_lba: logical block where backup partition array starts
        alternate_array_byte: byte where the backup partition array starts
    """

    def __init__(self, size: int, sector_size: int = 512) -> None:
        """Init Geometry with size in bytes"""
        self.sector_size = sector_size
        self.total_bytes = size
        self.total_sectors = int(size / self.sector_size)
        self.total_lba = int(size / self.sector_size)
        self.header_length = 92
        self.array_max_length = 128 * 128
        self.first_usable_lba = 34
        self.last_usable_lba = int(self.total_lba - 34)
        self.my_lba = 1
        self.primary_header_byte = int(self.my_lba * self.sector_size)
        self.partition_entry_lba = 2
        self.primary_array_byte = int(self.partition_entry_lba * self.sector_size)
        self.alternate_lba = int(self.total_lba - 1)
        self.alternate_header_byte = int(self.alternate_lba * self.sector_size)
        self.alternate_array_lba = int(self.total_lba - 33)
        self.alternate_array_byte = int(self.alternate_array_lba * self.sector_size)
